/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (collect.C)
  
  Copyright:  Damian J. J. Farnell

  Date:       Wed Mar 15 15:20:28 GMT 2000

  Updated:    Joerg Schulenburg 2004-2005

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: collect.C 

  PURPOSE: collect same configurations to reduce loops

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
// #include <math.h>
// #include <string.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"
// #include "nr.h"
// #include "nrutil.h"
#include "storage.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#ifndef DO_COLLECT
#define DO_COLLECT 1  /* set to 0 to switch off collect, default = 1 */
#endif

/* Joerg05:
 * who is calling  collect_term?
 *  62 times by obtain()+term2()..term62() in obtain.C 
 * parameters and its ranges?
 *  indices[0] - configuration (lattice animal), 1..NumberConfigurations
 *                                          new: 0 for GS Energy
 *  indices[1] - term 2..62, 999999=end (not used in solve!)
 *  indices[2..5] -  0,1..NumConfig ??  (0,0,0,0)=not_used(check it)
 * Outputfile fp1=*.terms is used for solve()
 * Outputfile fp1=*_gs    is used for ...     (new Sep05) ???
 * maximum number of terms and clusters for sample models?
 *  MAX_NUMBER_TERMS_IN_CCM_EQNS to big? (*46Bytes)
 * ToDo: dynamic arrays
 */

// MAX_NUMBER_TERMS_IN_CCM_EQNS renamed to MAXNUM_HASHED_TERMS
// MAXNUM_HASHED_TERMS should be a prime number for good hashing
// static arrays: MAXNUM_HASHED_TERMS * 44 byte

const int MAXNUM_HASHED_TERMS=(1<<19)-3; // 2^19-1 for s=1/2-LC-14   

#define MAX_out 200            // reduced 1000 to 200 Mar08

#if DO_COLLECT
// gpg --gen-prime 1 31
#define HASH_PRIME 0x701F35F3   // not used @ moment

/* export for measure speed */
long nterms_out=0; /* count output terms per thread (summed at end and stop output) */
long nterms_in=0;  /* count input  terms per thread (summed at end) */


// work: indices will be sorted (not used in the calling routine), Joerg Sep05
//       same or permutated indices[2..5] will be stored as sum only once
//       indices pointing to configurations 1..NumberConfigurations
//       permutated indices dont have the same probability! that means:
//       its difficult to simplify (no easy symmetry)
//       but it looks like that every term appears twice (ToDo: simplify!)
/* id: 0=gs_terms, 1=terms */
void collect_term(int indices[6], double dummy, int id) 
{
  int  k; long i, j; // , nio=7;
  static long na_in=0, na_out=0; /* summed over all bonds */
  static int  id_old=-1;         /* flag for new collection */
  static long counter=0;         /* buffer index for desciptors,multipliers */
  static long max_counter=0;     /* maximum buffer size */
  static int    descriptors[MAXNUM_HASHED_TERMS+1][6]; // 24 byte
  static double multipliers[MAXNUM_HASHED_TERMS+1];    //  8 byte
  // it is very slow to search a big table for an entry sequentialy
  // so we introduce a hash table of indizes to that entries
  //  and a table of indices to that table to have fast sequential access 
  //  to spare filled hash table (Joerg Sep05, xxzLC-14 5x faster)
  // number of hashed entries is in counter
  static long hash[MAXNUM_HASHED_TERMS+1]; // 8B hashed index to entries
  static int  hidx[MAXNUM_HASHED_TERMS+1]; // 4B index to hash entries
  static int hcol=0;  // counter for hash collisions, just as info
  static int hhit=0;  // counter for hash hits
  static int hnew=0;  // counter for new hash entries
  static int hval=-1;  // hash value
  static long numiterms[63], // count number of terms, only verbose output
              numoterms[63]; // ToDo: counting 1..4-fold-terms would be enough
  long maxterms_out=0, maxterms_in=0;


  if (nterms_in==0 && id_old != id) {  // new file
    for (i=0;i<63;i++) numiterms[i]=numoterms[i]=0; // initialize counter
  }

  nterms_in++;
  if (hval<0) { hval=0; // used as flag, initialize hash once!
    for (i=0;i<MAXNUM_HASHED_TERMS+1;i++) { hidx[i]=hash[i]=0; }
  }

  if (counter){  // something stored into buffer
      /* save collected terms and reset buffer for next term */

    if (    descriptors[counter][0] < indices[0]  /* next configuration number? */
//  Joerg Sep05: indices[1]=term not used in solve, so we can collect it too
       || descriptors[counter][1] < indices[1]  /* next term? 2..62 */
       ||                       0 > indices[0]  /* flush buffer condition */
      /* Sept 05: old was 0>i[0] but collect_gse needs 0>=i[0] */
       ||    NumberConfigurations < indices[0]
       || counter >= MAXNUM_HASHED_TERMS) /* buffer full */
    {
        if (       MAXNUM_HASHED_TERMS == counter) {
           static long nterms_out2=0;
           if (nterms_out2<10) {
             printf("MAXNUM_HASHED_TERMS is to small, flush buffer\n");
             printf("Overflow happens for cfg= %d  term= %d (%d %d %d %d)\n",
               indices[0],indices[1],
               indices[2],indices[3],indices[4],indices[5]);
             fflush(stdout);
           }
           if (++nterms_out2==10)
            printf("excessive output is stopped after %d prints\n",MAX_out);
        }
	// first save information 
	for(j=1;j<=counter;j++){
	  
          term_store(id,descriptors[j],multipliers[j]);
	  
	  // printf out information
	  
	  if (Control.output > 1 && myid == 0 && nterms_out < MAX_out) {
	    printf("OUT: Config=%3d : Term = %2d : CCM = (%3d,%3d,%3d,%3d) : "
		   "mult=% 7.3f\n",
		   descriptors[j][0], descriptors[j][1],
		   descriptors[j][2], descriptors[j][3],
                   descriptors[j][4], descriptors[j][5], multipliers[j]);
            if (nterms_out+1==MAX_out)
              printf("OUT: no output after MAX_out=%d terms\n", MAX_out);
          }
          nterms_out++;
          numoterms[descriptors[j][1]]++;
          hash[ hidx[j] ]=0; // reset hash entry
          hidx[j]=0; // not necessary
	  
	  // now overwrite the old stored information 
	  
	}
        for (i=1;i<=counter;i++) { hash[ hidx[i] ]=0; }
	counter=0;
      }
  }

  if (Control.output > 2 && myid == 0 && nterms_out < MAX_out)
    printf("IN:  Config=%3d : Term = %2d : CCM = (%3d,%3d,%3d,%3d)"
           " : mult=% 7.3f counter=%ld\n",
       indices[0], indices[1],
       indices[2], indices[3],
       indices[4], indices[5], dummy, counter);

  // the last config has lot of terms, try to reduce further by
  // reducing all possible permutations to one config by sorting (big first)

  for(;;) { // bubble sort indices
    j=1;
    for(i=2; i<5; i++)
      if (indices[i]<indices[i+1]) {  // exchange two indices
        j=0; k=indices[i+1]; indices[i+1]=indices[i]; indices[i]=k; 
      }
    if (j) break;
  }

  // Oct05: new hash table search (fast)
  j=0;
  if (indices[0]>=0 && indices[0]<=NumberConfigurations)
  {
    hval=(((((                          indices[5])
              *(NumberConfigurations+1)+indices[4])
              *(NumberConfigurations+1)+indices[3])
              *(NumberConfigurations+1)+indices[2])
              *(NumberConfigurations+1)+indices[0]);
    // try up to 1024 times,
    for(i=0;;i++) {
      hval = hval + indices[2] + 1;  // avoid overflow to negativ values
      // abs(hval) will fail for INT32_MIN
      if (hval<0) hval=~hval; // INT32_MIN..-1 <-> INT32_MAX..0
      hval %= MAXNUM_HASHED_TERMS; // MAX should be prime
      j = hash[hval];
      if (j>counter || j<0 || hval<0) { 
         printf("Bug86757 j=%ld counter=%ld hash=%d\n",j,counter,hval);
         exit(1);
      }
      if (j==0) {      // empty entry
             hash[hval]=counter+1;  // set hash entry
             hidx[counter+1]=hval;  // set sequential index to entry
             j=counter+1;           // store to table
             hnew++; break;
      } else {
        if ( indices[0]==descriptors[j][0] &&
    term_13[indices[1]]==term_13[descriptors[j][1]] &&  // term
             indices[2]==descriptors[j][2] &&
             indices[3]==descriptors[j][3] &&
             indices[4]==descriptors[j][4] &&
             indices[5]==descriptors[j][5]
           ) {
            multipliers[j]+=dummy;  // hashed == searched
            hhit++; break;
         } else if (i>=1024) {
           hcol++; nterms_out++;    // collision, write through
           numoterms[indices[1]]++;
           term_store(id,indices,dummy);
           if (hcol<20 && hcol>=0) { // hcol >= 0 avoids excessiv output
             if (!hcol) printf("%2d: Warning: collision counter overflow\n",myid);
             printf("%2d: collision hash[%d]=%ld lenhash=%ld=%.3f%%\n",
                     myid, hval, hash[hval], counter,
                     counter*1.0/MAXNUM_HASHED_TERMS);
             printf("IN:   Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d)\n",
                     indices[0],indices[1],
                     indices[2],indices[3],
                     indices[4],indices[5]);
             printf("HASH: Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d)\n",
                     descriptors[j][0],descriptors[j][1],
                     descriptors[j][2],descriptors[j][3],
                     descriptors[j][4],descriptors[j][5]);
           }
           j=0; break;
         }
      }
    }
  }
  
  if (indices[0]>=0 && indices[0]<=NumberConfigurations) {
    if (indices[1]<0 || indices[1]>62) { 
      printf("Error collect L%d: term=%d\n",__LINE__,indices[1]); exit(1); }
    numiterms[indices[1]]++; // for statistics only
  }

  if (indices[0]>=0 && indices[0]<=NumberConfigurations)
  if ( j>counter /* && indices[1] != 0 */ ) {
    counter++;  /* potential overflow! */
    if (counter>max_counter) max_counter=counter; /* just as info */
    descriptors[counter][0]=indices[0];
    descriptors[counter][1]=indices[1];
    descriptors[counter][2]=indices[2];
    descriptors[counter][3]=indices[3];
    descriptors[counter][4]=indices[4];
    descriptors[counter][5]=indices[5];
    multipliers[counter]=dummy;
  }

  if (indices[0]>NumberConfigurations || indices[0]<0) {
    term_store(id,indices,dummy);

    if (Control.output > 1 && myid == 0 ) {
      printf("OUT: Config=%3d : Term = %2d\n", indices[0], indices[1]);
    }
    if (Control.output > 0 /* && myid == 0 */ ) {
      /* ToDo: clock output here? */
      int t2, tmin, tmax, tsum;
      t2=seconds(); // differ for each thread (collect min,max)
      tmin=tmax=tsum=t2;
      maxterms_in =nterms_in;  /* set default for single thread computation */
      maxterms_out=nterms_out;
#if (PARALLEL_FLAG> 0)
      int tmp[63], i;
      long tmp2[63];
      MPI_Allreduce(numiterms, tmp2, 63, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      for(i=0;i<63;i++) numiterms[i]=tmp2[i];   // input terms
      MPI_Allreduce(numoterms, tmp2, 63, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      for(i=0;i<63;i++) numoterms[i]=tmp2[i];   // output terms
      if (myid>0) for(i=0;i<63;i++) numoterms[i]=numiterms[i]=0;
      MPI_Allreduce(    &t2, &tmin,  1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
      MPI_Allreduce(    &t2, &tmax,  1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
      MPI_Allreduce(    &t2, &tsum,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      MPI_Allreduce(  &hcol,   tmp,  1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      hcol=tmp[0];
      MPI_Allreduce( &nterms_in, tmp2, 1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
      maxterms_in=tmp2[0];
      MPI_Allreduce(&nterms_out, tmp2, 1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
      maxterms_out=tmp2[0];
      MPI_Allreduce( &nterms_in, tmp2, 1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      nterms_in=tmp2[0];
      MPI_Allreduce(&nterms_out, tmp2, 1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      nterms_out=tmp2[0];
      MPI_Allreduce(&max_counter,tmp2, 1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
      max_counter=tmp2[0];
#endif
      /* output debugging infos */
      if (id_old != id) { id_old=id; na_out=na_in=0; } // new file
      na_out+=nterms_out; na_in+=nterms_in-1;
      // output this as a statistic information
      if (myid==0) {
        long sumi=0, sumo=0;
        const char *term63s[63+1]={"",
                             " 1",
                             " 2 zz           Fm.S+m.Szk  [1]",
                             " 3 zz           Fk.S+k.Szm  [1]",
                             " 4 zz          Hkm.S+k.S+m  [1]",
                             " 5 zz        Fk.Fm.S+k.S+m  [2]",
                             " 6 z+              S+m.Szk  [0]",
                             " 7 z+          Fkl.S+k.S+m  [1]",
                             " 8 z-       -2*Fml.Szk.Szm  [1] cancels 17?",
                             " 9 z-      -2*Hkml.S+k.Szm  [1]",
                             "10 z-       -Fm.Fm.S+m.Szk  [2]",
                             "11 z-    -2*Hkm.Fm.S+k.S+m  [2]",
                             "12 z-     -2*Fk.Fm.S+k.Szm  [2]",
                             "13 z-    -Fk.Fm.Fm.S+k.S+m  [3]",
                             "14 +z              S+k.Szm  [0]",
                             "15 +z           Fm.S+m.S+k  [1]",
                             "16 -z      -2*Hmkl.S+m.Szk  [1]",
                             "17 -z       -2*Fml*Szk*Szm  [1]",
                             "18 -z     -2*Fm.Fk.S+m.Szk  [2]",
                             "19 -z    -2*Hkm.Fk.S+k.S+m  [2]",
                             "20 -z           -Fm.Fm.S+m  [2]",
                             "21 -z    -Fk.Fk.Fm.S+k.S+m  [3]",
                             "22 +-        -2*Fm.S+k.Szm  [1]",
                             "23 +-       -Fm.Fm.S+k.S+m  [2]",
                             "24 -+        -2*Fk.S+m.Szk  [1]",
                             "25 -+       -Fk.Fk.S+k.S+m  [2]",
                             "26 ++              S+k.S+m  [0]",
                             "27 --        4*Hkm.Szk.Szm  [1]",
                             "28 --    2*Hkm.Hkm.S+k.S+m  [2]",
                             "29 --     4*Hkm.Fm.S+m.Szk  [2]",
                             "30 --      4*Fk.Fm.Szk.Szm  [2]",
                             "31 --   2*Fm.Fm.Fk.S+m.Szk  [3]",
                             "32 --  4*Fk.Fm.Hkm.S+k.S+m  [3]",
                             "33 --   2*Fk.Fk.Fm.S+k.Szm  [3]",
                             "34 --  Fk.Fk.Fm.Fm.S+k.S+m  [4]",
                             "35 --     4*Hkm.Fk.S+k.Szm  [2]",
                             "36 --       2*Mkkm.S+k.Szm  [1] 2s>1",
                             "37 --   2*Mkmm.S+m.Szz  [1] NOTE: DIRECTIONALITY TO BE ADDED FOR TRIANGLE!",
                             "38 --    Nkkmm.S+k.S+m  [1]",
                             "39 --  Fm.Mkkm.S+k.S+m  [2]",
                             "40 -- 2*Fk.Hmm.S+m.Szk  [2]",
                             "41 --  Fk.Mmmk.S+k.S+m  [2]",
                             "42 -- 2*Fm.Hkk.S+k.Szm  [2]",
                             "43 --  Hkk.Hmm.S+k.S+m  [2]",
                             "44 --    Hkk.Fm.Fm.S+k.S+m  [3] 2s>1",
                             "45 --    Hmm.Fk.Fk.S+k.S+m  [3] 2s>1",
                             "46 z-    -Hmm.S+m.Szk  [1] 2s>1",
                             "47 z-   -Mkmm.S+k.S+m  [1] 2s>1",
                             "48 z- -Fk.Hmm.S+k.S+m  [2] 2s>1",
                             "49 -z   -Mkkm.S+k.S+m  [1] 2s>1",
                             "50 -z    -Hkk.S+k.Szm  [1] 2s>1",
                             "51 -z -Fm.Hkk.S+k.S+m  [2] 2s>1",
                             "52 +-    -Hmm.S+k.S+m  [1] 2s>1",
                             "53 -+    -Hkk.S+k.S+m  [1] 2s>1",
                             "54 z           Fk.S+k  [1]",
                             "55 -        -2*Fk.Szk  [1]",
                             "56 -         -Hkk.S+k  [1]",
                             "57 -       -Fk.Fk.S+k  [2]",
                             "58 z^2    2Fk.S+k.Szk  [1]",
                             "59          Fk.S+k.S+k [1]",
                             "60        Hkk.S+k.S+k  [1]",
                             "61      Fk.Fk.S+k.S+k  [2]",
                             "62 +              S+k  [0]",
                             ""};
        for (i=0;i<63;i++){
          if(numiterms[i]) printf(" num_terms%02ld: %12ld => %12ld  %s\n",
                                  i, numiterms[i], numoterms[i], term63s[i]);
          sumi+=numiterms[i];
          sumo+=numoterms[i];
        }
        printf(" collect_term maxbuff= %8ld\n", max_counter);
        printf(" collect_term hash collisions= %6d\n", hcol);
        printf(" collect_term maximum unbalance in= %6.2f%% out= %6.2f%%"
               " t=%d..%d..%ds npe= %d\n",
                maxterms_in *100.*npe/nterms_in -100.,
                maxterms_out*100.*npe/nterms_out-100., tmin, tsum/npe, tmax, npe);
        printf(" collect_term bond summed terms in= %12ld  out= %10ld\n", sumi,  sumo);
        printf(" collect_term bond summed terms in= %12ld  out= %10ld\n", na_in, na_out);
        printf(" collect_term storage memory per term: 6*int=%d\n", 6*sizeof(int));
      }
    }
    max_counter=0; nterms_out=nterms_in=0;
    hnew=hhit=hcol=0;   // reset number of collisions
  }
}
#else /* dont collect */
void collect_term(int indices[6],double dummy, int id) 
{
  int nio;

  term_store(id,indices,dummy);
	  
  if (Control.output > 1 && myid == 0 && nterms_out < MAX_out) {
     printf("OUT: Config=%3d : Term = %2d : CCM = (%3d,%3d,%3d,%3d) : "
     "mult=%10.7f\n",
     indices[0],indices[1],
     indices[2],indices[3],
     indices[4],indices[5],
     dummy);
     if (nterms_out+1==MAX_out)
        printf("excessive output is stopped after %d terms\n", MAX_out);
  }
  if (indices[0]>NumberConfigurations || indices[0]<0) {
    /* all terms collected, obtain finished for one bond */
#if (PARALLEL_FLAG> 0)
      long tmp;
      MPI_Allreduce(&na_out,  &tmp,  1, MPI_LONG, MPI_SUM, MPI_COMM_WORLD);
      na_out=tmp;
#endif
    if (Control.output > 0 && myid == 0 ) {
      /* output debugging infos */
      printf(" collect_term  terms out= %8ld\n", nterms_out);
      nterms_out=0;
    }
  }
  nterms_out++;
  nterms_in++; /* same, just for compability */
}
#endif
